/******************************************************************\
**                                                                **
**     GL1.js : (basic) emulation of OpenGL 1.x in webGL :        **
**                                                                **
**           - immediate Mode ( glBegin / glVertex3f / glEnd )    **
**           - gl & glu matrices, light & materials,              **
**           - display lists, picking, wire-mode,                 **
**           - 1D/2D-textures , texGen,                           **
**           - glut events and menus,                             **
**           - glu/glut goodies: usual solid objects, fonts       **
**                                                                **
**    use and extend webgl, use gl-matrix, contextMenu            **
**                              ( + require, jquery )             **
**                                                                **
**    version 4.2             Fabrice NEYRET , september 2015     **
**                                                                **
\******************************************************************/
/*

DISCLAIMER:
  - This lib was first intended only to port to Web my very old OpenGL 1.0 demos.
  - I learned JavaScript and WebGL while writing it.
  -> so, don't expect exemplar coding nor full implementation of everything !
  -> I haven't tested all combinatories and so basic bugs probably remain.

  I hope it might help as such, or as a first basis,
  or at least help people transitioning from old friendly to new elitist and unconcise OpenGL....
  

* Implemented (basically) : ---

- draw in immediate mode :
  
    gl.Begin()/glEnd() with gl.POINTS,
                            gl.LINE/S/STRIP/LOOP,
                            gl.TRIANGLE/S/STRIP/FAN,
			    gl.QUAD/S/STRIP,   gl.POLYGON ( <- welcome back ! )
    gl.vertex2/3f[v](), gl.color3/4f[v](), gl.normal3f[v](), gl.texCoord2f[v]()

    gl.polygonMode()                  [ ignore per-side  ]
    gl.pointSize()
    gl.push/popAttrib()               [ limited to color ]

- display lists:
   gl.genLists(), gl.newList(), gl.endList(), gl.callList()
   [ minimal support: only 1 list, no lights in lists,
                      no change of shadeModel or texGen ]


- glu & glut solids:

   glu.sphere(), glu.disk(), glu.cylinder()
   glut.*Sphere(), glut.*Cone(), glut.*Torus(), glut.*Cube(),  
   glut.*Tetrahedron(), glut.*Icosahedron(),     with * = solid , wire

- glut characters                     [ strokes Roman only ]
   glut.strokeCharacter(), glut.strokeString(),
   glut.strokeWidth[f](), glut.strokeLength[f](), glut.strokeHeight()


- matrices :
  
    gl.matrixMode()  with  gl.PROJECTION, gl.MODELVIEW, gl.TEXTURE, gl.COLOR
                                      [ but only the 2 first are used ]
    gl.pushMatrix(), gl.popMatrix()
    gl.loadIdentity(), gl.ortho(), gl.frustum(), glu.perspective(), glu.lookAt()
    gl.translatef(), gl.scalef(),  gl.rotatef(), glMultMatrixf();
    glu.unProjectv [ API return value: last parameters were &f,&f,&f -> now vector]


- lighting functions : 
  
    gl.shadeModel()            
    gl.lightfv()               [ only LIGHT0 (auto-set) ]
    gl.colorMaterial()         [ set to gl.AMBIENT_AND_DIFFUSE, 2-sided material limited to constant back color ]
    glMaterialfv()             [ thus ignored ]
    gl.fogf[v]()               [ basic support: GL_EXP, no START/END ]


- textures:

   gl.genTextures(), gl.tex[Sub]Image1/2D(), glu.build1/2DMipmaps()
                [ NB: the regular webGL teximage2D func(image) is renamed gl.texDOMImage2D ]
   gl.texEnv*()  [ minimal support: ENV_MODE ]
   gl.texGen*()  [ minimal support: OBJECT_LINEAR / OBJECT_PLANE ]

- picking:

    glu.pickMatrix(), gl.loadName(), gl.pushName() [stack forced at 1]
    gl.selectBuffer(), gl.renderMode()  [ only SELECT mode, returned z not set ]
    [ take care: antialiasing corrupt labels. -> use thick primits, big point size]


- glut windowing :
 
    glut.createWindow()
    glut.get()
    glut.init()                       [ do nothing ]
    glut.initDisplayMode()            [ do nothing ]
    glut.initDisplayString()          [ do nothing ]
    glut.initWindowSize()             [ see canvas size issue ]
    glut.mainLoop()                   [ do nothing ]
    glut.swapBuffers() 
    gl.drawBuffer(), gl.readBuffer()  [ do nothing ]
    glut.reshapeWindow()              [ see canvas size issue ]
    glut.positionWindow()             [ see canvas size issue ]
  

- glut events and callBacks :
  
    glut.postRedisplay(), glut.fullScreen()         
    glut.idleFunc()              
    glut.displayFunc(), glut.reshapeFunc()
    glut.keyboardFunc(), glut.specialFunc(), glut.keyboardUpFunc(), glut.ignoreKeyRepeat()
    glut.mouseFunc(), glut.motionFunc()
    glut.visibilityFunc()          [ not connected ]

- glut menus :
    glut.createMenu(), glut.addMenuEntry(), glut.addSubMenu(),
     glutSetMenu(), glut.changeToMenuEntry()
    glut.attachMenu(), glut.detachMenu() 


- plus all necessary flags, and extension of gl.enable/disable()
- plus glGet* functions   [ for scalars, API return value in last parameters -> now as returned value ]

- if a html element "fps" exists, framerate is written in it.
- if a html element "stdout" exists, printf() writes in it. Visibility is switched on/off when the key mentionned in triggerKey attribute is hit.
  sprintf() is also implemented [ but string is returned, not through 1st arg]
  
      
* Issues: ---
  
  - canvas layout  (full navigator window or full screen, window unchanged from program). 
  

* To be implemented: ---

  - completed material, fonts, menus
  - 4f for vertex, normal, textures, light
  - glut multi/subwindows,
  - perfs, cleanup ( especially display list pile of ctx )
  - tests ! 
*/




// these 3 classes emulate OpenGL 1 contextes
this.glut = {};
this.glu  = {};
this.gl; // set in glut.createWindow()


var glMatrix = require('./gl-matrix'); // https://github.com/toji/gl-matrix
require('./glutstrokefonts'); // ported from freeglut http://freeglut.sourceforge.net/

// contextMenu.js   http://ignitersworld.com/lab/contextMenu.html
require('./jquery.js'); 
require('./contextMenu.js');
require('./contextMenuCSS.js'); // ugly, but how to get the right path otherwise ?
$("head").append('<style>'+menuCSS+'</style>');
// $("head").append( $("<link rel='stylesheet' type='text/css' href='./lib/contextMenu.css'>" )); 




// --- context-independent functions. --------------------------------
// needed before createWindow initializations.

glut.init              = function ()     {};
glut.initDisplayMode   = function (mode) {};
glut.initDisplayString = function (mode) {};
glut.initWindowSize    = function (l,h)  {};  // should do something

// --- multi-windows
/*   
   var glutWindows = { index : 0; curWindow : -1; gluts = {} };
   glut.GetWindow = function() { return glutWindows.curWindow; }
   glut.SetWindow = function( w ) {
     curWindow = w; glut = gluts[w];
     }
*/

// --- createWindow + init all the other functions --------------------

glut.createWindow  = function (canvas) { 

    var gl1 = {}; // our extended gl context

    gl = initInterface(canvas);  // window and events (mostly glut operators)
    initGL();                    // OpenGL1 operators (begin/end/vertex...)
    initMatrices();              // matrices objects and operators
    initLighting();              // lighting operators
    initShaders();               // init shaders, and their connectors to data

    gl1.initShaders = initShaders;

    return;

        

    
    // --- init shaders, and their connectors  --------

    function initShaders() {
	//gl1.initShaders = this;
	
	// --- default pixel & fragment shader

	// NB: these are minimal shaders: 1 light (at infinity), 1 texture, simple material
        // (with fog, 2-sided and texture-blend modes, though).
	// if you want more, see http://www.gamedev.net/page/resources/_/technical/opengl/creating-a-glsl-library-r2428 or https://en.wikibooks.org/wiki/GLSL_Programming/GLUT


	// Would it be more efficient to specialize shaders depending of features used ?
	// attention: variable unsused in the shader are suppressed at compilation.

	var shader_vs = // <script id="shader-vs" type="x-shader/x-vertex">
` precision mediump float;

  attribute vec3 aVertex;
  attribute vec3 aNormal;
  attribute vec2 aTexCoord;
  attribute vec4 aColor;
  attribute vec4 aColorB;

  varying vec3 vPosition;
  varying vec4 vColor;
  varying vec4 vColorB;
  varying vec3 vNormal;
  varying vec2 vTexCoord;
    
  uniform mat4 uMVMatrix;
  uniform mat4 uPMatrix;
  uniform mat4 uNMatrix;
  uniform float uPointSize;

 void main(void) {
     gl_Position = uPMatrix * uMVMatrix * vec4(aVertex, 1.);
     vPosition = gl_Position.xyz;
     vNormal   = ( uNMatrix * vec4(aNormal, 0) ).xyz;
     vColor    = aColor;
     vColorB   = aColorB;
     vTexCoord = aTexCoord;
     gl_PointSize = uPointSize;
  }
`;

	var shader_fs = // <script id="shader-vs" type="x-shader/x-fragment">
` precision mediump float;

  varying vec3 vPosition;
  varying vec4 vColor;  
  varying vec4 vColorB;
  varying vec3 vNormal;
  varying vec2 vTexCoord;

  uniform vec4 light0Position; // light0 assumed at infinity
  uniform vec4 light0Ambient;
  uniform vec4 light0Diffuse;
  uniform vec4 light0Specular;
  // material ambiant&diffused assumed in  col
  uniform vec4 materialSpecular;
  uniform float materialShininess;
  uniform vec4 materialEmission;

  uniform float fogDensity;
  uniform vec4 fogColor;
	
  uniform int lightingEnabled; 
  uniform int fogEnabled;

  uniform int TEXTURE_ENV_MODE;
  uniform sampler2D uSampler;
	
void main(void) {

    // material color
    vec4 color = (vColorB.r>=0. && !gl_FrontFacing) ? vColorB : vColor;

// flags GL_ from gl.h
#define TEX_REPLACE		0x1E01
#define TEX_MODULATE		0x2100
#define TEX_DECAL		0x2101
#define TEX_BLEND		0x0BE2
#define TEX_ADD			0x0104

    // texturing
    if (TEXTURE_ENV_MODE > 0) {
	vec4 colorT = texture2D(uSampler, vec2(vTexCoord.s, vTexCoord.t));
	if      (TEXTURE_ENV_MODE==TEX_REPLACE)
	{ color.rgb  = colorT.rgb; if (true/*RGBA*/) color.a = colorT.a; } 
	else if (TEXTURE_ENV_MODE==TEX_MODULATE) color *= colorT;
	else if (TEXTURE_ENV_MODE==TEX_ADD)
	{ color.rgb += colorT.rgb; color.a *= colorT.a; }
	else if (TEXTURE_ENV_MODE==TEX_DECAL)
	{ color.rgb = mix(color.rgb,colorT.rgb,color.a); color.a *= colorT.a; }
    }

    // Phong shading  - 1 light, no attenuation
    if (lightingEnabled > 0) {
	vec3 n = normalize(vNormal), l = normalize(light0Position.xyz);
        float diffuse =   max( 0., dot(n,l) );
        color.rgb *=  ( diffuse*light0Diffuse + light0Ambient ).rgb; // assume col = materialAmbient&Diffuse
	if (any(bvec3(materialSpecular.xyz))) {
	    vec3 h = normalize(vec3(0,0,1)+l);
	    float specular = pow( max( 0., dot(n,h) ) , materialShininess);
	    color.rgb += ( specular*light0Specular*materialSpecular ).rgb;
	}
	color.rgb += materialEmission.rgb;
     }

    // fog
    if (fogEnabled > 0) {
        float transp = exp(-vPosition.z*fogDensity);
        color = mix(fogColor, color, transp);
    }

    //color = vec4(0); /*light0Ambient;*/ color.a=.20; // <<

    gl_FragColor = color; 
  }
`;
	// --- set up shaders
	
	var fragmentShader = getShader(gl, shader_fs, gl.FRAGMENT_SHADER);
	var vertexShader   = getShader(gl, shader_vs, gl.VERTEX_SHADER);
      
	var shaderProgram = gl.createProgram();
	gl1.shaderProgram = shaderProgram;
	gl.attachShader(shaderProgram, vertexShader);
	gl.attachShader(shaderProgram, fragmentShader);
	gl.linkProgram(shaderProgram);
	
	if (!gl.getProgramParameter(shaderProgram, gl.LINK_STATUS)) 
	    alert("Could not initialise shaders");
	
	gl.useProgram(shaderProgram);

	// --- set connectors

	// vertices and attributes
	shaderProgram.vertexPositionAttribute = gl.getAttribLocation(shaderProgram, "aVertex");
	gl.enableVertexAttribArray(shaderProgram.vertexPositionAttribute);
		
	shaderProgram.colorAttribute = gl.getAttribLocation(shaderProgram, "aColor");
	gl1.colorEnabled = (shaderProgram.colorAttribute != null );
	//if()   gl.enableVertexAttribArray(shaderProgram.colorAttribute);
	shaderProgram.colorBAttribute = gl.getAttribLocation(shaderProgram, "aColorB"); // limited management of 2-sided materials
	
	shaderProgram.normalAttribute = gl.getAttribLocation(shaderProgram, "aNormal");
	gl1.normalEnabled = (shaderProgram.normalAttribute != null );
	//if()   gl.enableVertexAttribArray(shaderProgram.normalAttribute);
	
	shaderProgram.texCoordAttribute = gl.getAttribLocation(shaderProgram, "aTexCoord");
	gl1.texCoordEnabled = (shaderProgram.texCoordAttribute != null );
	//if()   gl.enableVertexAttribArray(shaderProgram.texCoordAttribute);

	// matrices
	shaderProgram.pMatrixUniform = gl.getUniformLocation(shaderProgram, "uPMatrix");
	shaderProgram.mvMatrixUniform = gl.getUniformLocation(shaderProgram, "uMVMatrix");
	shaderProgram.nMatrixUniform = gl.getUniformLocation(shaderProgram, "uNMatrix");

	// light, fog, material
	shaderProgram.lightingEnabled = gl.getUniformLocation(shaderProgram, "lightingEnabled");
	shaderProgram.light0PositionUniform = gl.getUniformLocation(shaderProgram, "light0Position");
	shaderProgram.light0AmbientUniform = gl.getUniformLocation(shaderProgram, "light0Ambient");
	shaderProgram.light0DiffuseUniform = gl.getUniformLocation(shaderProgram, "light0Diffuse");
	shaderProgram.light0SpecularUniform = gl.getUniformLocation(shaderProgram, "light0Specular");

	shaderProgram.materialSpecularUniform = gl.getUniformLocation(shaderProgram, "materialSpecular");
	shaderProgram.materialShininessUniform = gl.getUniformLocation(shaderProgram, "materialShininess");
	shaderProgram.materialEmissionUniform = gl.getUniformLocation(shaderProgram, "materialEmission");

	shaderProgram.fogEnabled = gl.getUniformLocation(shaderProgram, "fogEnabled");
	shaderProgram.fogColorUniform = gl.getUniformLocation(shaderProgram, "fogColor");
	shaderProgram.fogDensityUniform = gl.getUniformLocation(shaderProgram, "fogDensity");

	// misc
	shaderProgram.pointSizeUniform = gl.getUniformLocation(shaderProgram, "uPointSize");
	shaderProgram.textureSamplerUniform = gl.getUniformLocation(shaderProgram, "uSampler");
	shaderProgram.textureEnvModeUniform = gl.getUniformLocation(shaderProgram, "TEXTURE_ENV_MODE");
	

    
	function getShader(gl, str, type) {
	    var shader = gl.createShader(type); // replace above
	    gl.shaderSource(shader, str);
	    gl.compileShader(shader);
	    
	    if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
		alert(gl.getShaderInfoLog(shader));
		return null;
	    }	
	    return shader;
	}
    }



    

    // ---  emulates OpenGL 1 immediate mode ( glBegin/glVertex3f/glEnd ) --------

    function initGL() {
     	vec4 = glMatrix.vec4;

	// Primitives --- missings from from gl.h
	gl.QUADS	  =	0x0007;
	gl.QUAD_STRIP     =	0x0008;
	gl.POLYGON	  =	0x0009;
	
	// arrays & indexes defined & (re)initialized in gl.begin	
	gl1.vertexArray  = [ ]; 
	gl1.normalArray  = [ ];
	gl1.colorArray   = [ ];
	gl1.texCoordArray= [ ];
	gl1.indexArray   = [ ];
	gl1.index=0; // , type, first, second;

	
	// current state for normal, colors, uv. side effects welcomed :-)
	gl1.nx=0,  gl1.ny=0,  gl1.nz=1,
	gl1.r=1,   gl1.g=1,   gl1.b=1,   gl1.a=1,
	gl1.rB=-1, gl1.gB=-1, gl1.bB=-1, gl1.aB=-1,
	gl1.u=0,   gl1.v=0; 

	gl1.INDEXED = false;
	gl1.inDraw = false;
	gl1.ctx = {};   // for materials, etc. init in init_lighting
	  // should really contain also current color, etc.
          // issue: in display-lists, when gl.begin/end only has some gl.color.

	// --- display lists
	// todo: immediate transforms for glut.tetra/icosa 
        //       retarded transform for lightpos
	
	gl1.proto = {}; // for data shared by all instances of gl1 (list elements)
	// what is NOT shared in case of lists:
	// arrays, index, type, cur_attributes if isSet[0], some states.
	// What we should do: put all this in ctx, stack ctxs, keep gl1 unique, add a proto_ctx = ctx (no list) or = merged state (drax list).
	gl1.proto.inList = 0; //  0, -1 or 1 during list def
	gl1.proto.callList = false; // during callList
	gl1.proto.elements = [];    // stack of gl1s per begin/end block
	gl1.proto.nbElements = 0;
	gl1.proto.ctx = gl1.ctx;   // at callList, current merge of up and loc
	// to merge material states during callList eval
	gl1.proto.ctx.merge = function(ctx)
	                       { for (var x in ctx) this[x]=ctx[x]; }
	// to propagate local state to next  begin/end block at construction
	// (indeed there are too many things in there).
	gl1.clone = function() { var O={}; for (var x in this) O[x]=this[x]; return O;}

	//  from gl.h
	gl.COMPILE =		  0x1300;
	gl.COMPILE_AND_EXECUTE =  0x1301;

	gl.genLists = function(n) { return 1;}; // limited to 1 list
	
	gl.newList = function(list, mode) {
	    switch (mode) {
	    case gl.COMPILE_AND_EXECUTE: gl1.proto.inList = -1; break;
	    case gl.COMPILE:             gl1.proto.inList =  1; break;
	    }
	    gl1.proto.elements = [];
	    gl1.proto.nbElements = 0;

	    gl.pushMatrix();
	    gl.loadIdentity(); // from now, vertices and normals are directly transformed
	    //gl1.proto.ctx = gl1.ctx; // (already) save context materials
	    gl1.proto.gl1_up = gl1; // remaining states to save. Which, exactly ?
	               // first matter is to not spoil 1st & last saved states
	    gl1 = gl1.clone();
	    gl1.ctx = {}; // from now, the context only store new definitions
                          // Will be merged with new up_context at callList.
	}

	gl.endList = function() {
	    var mode = gl1.proto.inList; // to fit the case "compile & execute"
	    if (mode==0) return;    // (debug) we were not in a list !
	    gl1 = gl1.proto.gl1_up; // restore previous context
	    gl1.proto.inList = 0; 
	    gl.popMatrix();
	    // free all data ? or let garbage collector do ?

	    gl1.ctx = gl1.proto.ctx; // restore merged context. ~Redundant with proto.gl1.ctx
	    if (mode==-1) gl.callList(); // case "compile & execute"
	}

	gl.callList = function (list) { 
	    gl1.proto.callList = true;
	    // color up-state saved in proto 
	    gl1.proto.r =gl1.r  ; gl1.proto.g =gl1.g  ; gl1.proto.b =gl1.b ;  gl1.proto.a =gl1.a ;
	    gl1.proto.rB=gl1.rB ; gl1.proto.gB=gl1.gB ; gl1.proto.bB=gl1.bB;  gl1.proto.aB =gl1.aB ;
	    var gl1_ = gl1; // save current up-context;

	    for (i in gl1.proto.elements) { 
		gl1 = gl1.proto.elements[i]; 
		// inherit context if no default color set in list element
		gl1.proto.ctx.merge(gl1.ctx);
		// very ugly: separate merging for default color.
		// Current colors can't be in ctx due to different behavior 
                // depending of begin/end define none, some or all colors.
		if (gl1.ctx.colorSet0) 
		     { gl1.proto.r=gl1.r ; gl1.proto.g=gl1.g ; gl1.proto.b=gl1.b ;   gl1.proto.a =gl1.a ;}
		else { gl1.r=gl1.proto.r ; gl1.g=gl1.proto.g ; gl1.b=gl1.proto.b ;   gl1.a = gl1.proto.a;}
		if (gl1.ctx.colorBSet0) 
		     { gl1.proto.rB=gl1.rB ; gl1.proto.gB=gl1.gB ; gl1.proto.bB=gl1.bB;   gl1.proto.aB =gl1.aB ;}
		else { gl1.rB=gl1.proto.rB ; gl1.gB=gl1.proto.gB ; gl1.bB=gl1.proto.bB ;   gl1.aB = gl1.proto.aB ;}
		
		gl1.drawVBO(); 
	    }
	    gl1 = gl1_; //restore up-context
	    gl1.proto.callList = false;
	}
	
	// --- drawing 

	gl.begin = function (type) { // ------------------------
	    if (gl1.proto.inList) { // TODO: direct indexed mode (icosa/tet)
		if (gl1.proto.nbElements>0) { 
		    gl1 = gl1.clone(); // to keep the arrays and states (prepare push)
		    // first gl1.ctx has been reseted (in-list materials def)
		}
		gl1.calcNMatrix(); // in case MV has changed
	    }

	    gl1.inDraw = true; // in begin-end session
	    gl1.type = type;
            gl1.index  = 0;
	    gl1.first  = -1;
	    gl1.second = -1;
	    
	    gl1.vertexArray  = [ ]; // should be new Float32Array();
	    gl1.normalArray  = [ ]; // but these are not resizable :-(
	    gl1.colorArray   = [ ];
	    gl1.texCoordArray= [ ];
	    gl1.indexArray   = [ ];


	    gl1.normalSet   = false; // to check if defined in the session
	    gl1.colorSet    = false;
	    gl1.texCoordSet = false;

	    gl1.INDEXED = (gl1.shadeModel==gl.FLAT); // force sparse triangles
	                                 // not compatible with display-lists
	};

	//                             ---------------------------------
	gl.normal3f = function(nx,ny,nz) {
	    gl1.nx = nx; gl1.ny = ny; gl1.nz = nz;
	    if (gl1.inDraw)  gl1.normalSet = true; // normalArray used
	};
	gl.color4f  = function(r,g,b,a)    {
	    if ( gl1.side==gl.BACK) {
		// limited support of back-side material
		gl1.rB  = r;  gl1.gB  = g;  gl1.bB  = b; gl1.aB = a;
		// recognized via rB >= 0.
		gl1.ctx.colorBSet0 = true;
		return;
	    }
	    gl1.r  = r;  gl1.g  = g;  gl1.b  = b; gl1.a = a;
	    if (gl1.inDraw)  gl1.colorSet  = true; // colorArray used
	    gl1.ctx.colorSet0 = true; // (lists): color defined since the last end -> state set, to be kept in ctx stack. TODO: same for all attributes.
	};
	gl.texCoord2f= function(u,v)     {
	    gl1.u  = u;  gl1.v  = v;
	    if (gl1.inDraw)  gl1.texCoordSet = true;
	};

	gl.vertex3f    = function (x,y,z){ gl.vertex3fv([x,y,z]);          }
	gl.normal3fv   = function (n)    { gl.normal3f(n[0],n[1],n[2]);    }
	gl.color4fv    = function (c)    { gl.color4f(c[0],c[1],c[2],c[3]);}
	gl.color3fv    = function (c)    { gl.color4f(c[0],c[1],c[2],1.);  }
	gl.color3f     = function (r,g,b){ gl.color4f(r,g,b,1.);           }
	gl.texCoord2fv = function (t)    { gl.texCoord2f(t[0],t[1]);       }
        gl.vertex2f    = function (x,y)  { gl.vertex3fv([x,y,0.]) ;        }
	gl.vertex2fv   = function (p)    { gl.vertex3fv([p[0],p[1],0.]);   }
	gl.pushAttrib  = function (flag) { gl1.C = [gl1.r,gl1.g,gl1.b,gl1.a]; }
	gl.popAttrib   = function (flag) { gl.color4fv(gl1.C);             }
	
	// picking with select buffer
	/* Render Mode --- from gl.h */
	gl.FEEDBACK	= 0x1C01;
	gl.RENDER	= 0x1C00;
	gl.SELECT	= 0x1C02;
	gl.loadName    = function (id)     {
	    gl.color4f(Math.floor(id/256)/256., (id % 256)/256., 0,1);
	}
	gl.pushName    = function (id)     { gl.color4f(id,id,id,1.);  }
	gl.selectBuffer= function (n,buff) { gl1.selectBuffer = buff;  }
	// glu.pickMatrix def in initMatrices

	gl.renderMode  = function (flag) {
	    var nb = 0;
	    
	    if (gl1.renderMode == gl.SELECT) {
		// capture pick-rendered (colors=labels), 
		var x = gl1.pickViewport[0], y = gl1.pickViewport[1],
		    w = gl1.pickViewport[2], h = gl1.pickViewport[3];
		// glut.swapBuffers(); // to prevent bg blending
		var pixels = new Uint8Array(4*w*h); 
		gl.pixelStorei(gl.PACK_ALIGNMENT,1); //or 4  
		gl.pixelStorei(gl.UNPACK_ALIGNMENT,1); //or 4  
 		gl.readPixels(x-w/2, y-h/2, w, h, gl.RGBA, gl.UNSIGNED_BYTE, pixels);
		/* 
		printf("\n");
		for (var j=0; j<h; j++){
		    for (var i=4*w*j; i<4*w*(j+1); i+=4)
		    printf("%d %d %d %d   ",pixels[i],pixels[i+1],pixels[i+2],pixels[i+3]);
		    printf("\n");}
		*/
		var buff = [];
		for (var i=0; i<4*w*h; i+=4)
		    buff.push(pixels[i]*256+pixels[i+1]);
		// make histogramm of labels in select buffer
		buff = buff.sort(function(a, b){ return -(a - b); });
		gl1.selectBuffer.length=0;
		var c=-1, nb=0; 
		for (var i=0; i<w*h; i++)
		    if (buff[i]!=c) {
			c=buff[i], nb++;
			gl1.selectBuffer.push(1,0,1,c);
		    }
	    }
	    
	    // modifies projection matrix and viewport accordingly.
	    if (flag == gl.SELECT) {
		var x = gl1.pickViewport[0], y = gl1.pickViewport[1],
	            w = gl1.pickViewport[2], h = gl1.pickViewport[3];
		gl.viewport(x-w/2,y-h/2,w,h);
		gl1.applyPickMatrix(gl1.pickMatrix);
		// gl.disable(gl.BLEND);      
		// gl.disable(gl.DEPTH_TEST);  
		gl.clearColor(0,0,0,1);   
		gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT); 

	    }
	    else
		gl.viewport(0,0,gl1.width,gl1.height);
	    
	    gl1.renderMode = flag;
	    // return nb selected items
	    return nb;
	}
	
	gl.vertex3fv = function (p)    { // -------------------------

	    if (gl1.texGen) {
		if (gl1.texGenMode[0]==gl.OBJECT_LINEAR) {
		    gl1.u = vec4.dot([p[0],p[1],p[2],1],gl1.texGenPlane[0]);
		    gl1.texCoordSet = true;
		}
		if (gl1.texGenMode[1]==gl.OBJECT_LINEAR) {
		    gl1.v = vec4.dot([p[0],p[1],p[2],1],gl1.texGenPlane[1]);
		    gl1.texCoordSet = true;
		}
		
	    }

	    var n = [gl1.nx,gl1.ny,gl1.nz];
	    if (gl1.proto.inList) {
		// in a display-list, local transforms are applied directly
		// Maybe should be done at buffering in drawVBO
		// Todo: idem for indexed VBO
		var v= gl1.applyMV ([p[0],p[1],p[2],1]);
		p = [v[0],v[1],v[2]];
		v =  gl1.applyNMV([n[0],n[1],n[2],0]);
		n = [v[0],v[1],v[2]];
		// how is it possible that Float32Array.slice() not always implemented ???
		//p = gl1.applyMV ([p[0],p[1],p[2],1]).slice(0,3);
		//n = gl1.applyNMV([n[0],n[1],n[2],0]).slice(0,3);
	    } 

	    gl1.vertexArray.push([p[0],p[1],p[2]]);
	    if (gl1.normalEnabled)   gl1.normalArray.push ([n[0],n[1],n[2]]);
	    if (gl1.colorEnabled)    gl1.colorArray.push  ([gl1.r, gl1.g, gl1.b,gl1.a]);
	    if (gl1.texCoordEnabled) gl1.texCoordArray.push([gl1.u, gl1.v]);

	    if (gl1.INDEXED) { // if must output a indexArray 
		switch(gl1.type) {
		case gl.POINTS:
		case gl.LINES:
		case gl.TRIANGLES:
	     // case gl.QUADS:   // no longer exists (see below)
	     // case gl.POLYGON: // no longer exists
		gl1.indexArray.push(gl1.index); break;
		case gl.LINE_STRIP:
		case gl.LINE_LOOP:
		    if (gl1.first>=0)  gl1.indexArray.push(gl1.first, gl1.index);
		    else gl1.second = gl1.index;
		    gl1.first = gl1.index;
		    break;
		case gl.QUAD_STRIP: // no longer exists: emulation
		case gl.TRIANGLE_STRIP:
		    if (gl1.first>=0)
			if (gl1.first%2==0)
			     gl1.indexArray.push(gl1.first, gl1.second, gl1.index);
		        else gl1.indexArray.push(gl1.second, gl1.first, gl1.index);
		    gl1.first = gl1.second; gl1.second = gl1.index;
		    break;
		case gl.QUADS: // no longer exists: emulation
		    if ( gl1.index % 4 ==0) gl1.first = gl1.second = -1;
		    //if (gl1.polygonMode == gl.LINE) {
		    //   gl1.indexArray.push(gl1.index); break;}
		case gl.POLYGON: // no longer exists: emulation
		case gl.TRIANGLE_FAN:
		    if (gl1.first>=0)  
			gl1.indexArray.push(gl1.first, gl1.second, gl1.index);
		    else gl1.first = gl1.second;
		    gl1.second = gl1.index;
		    break;	      
		}
	    }
	    gl1.index++;
	    
	    // multiple quads are treated as multiple triangle fans
	    // (new primit after each, and by-pass flattening if contouring)
	    if (gl1.type==gl.QUADS &&  gl1.index==4) {
		//if (gl1.proto.ctx.polygonMode == gl.LINE) // not compatible with displayLists 
		//    gl1.type = gl.LINE_STRIP; 
		gl.end();
		gl.begin(gl.QUADS);
	    }
	};
     
	gl.end = function () { // ----------------------------------
	    // set gl1.INDEXED=true if you want to refer to your indexBuffer
	    
	    switch (gl1.type) {
	    case gl.QUAD_STRIP: //  no longer exists 
		gl1.type = gl.TRIANGLE_STRIP;
		break;
	    case gl.POLYGON: //  no longer exists
		gl1.type = gl.TRIANGLE_FAN; // not good for N quads
		break;
	    case gl.POINTS:
	    case gl.LINES:
	    case gl.LINE_STRIP:
	    case gl.TRIANGLES:
	    case gl.TRIANGLE_STRIP:
	    case gl.TRIANGLE_FAN:
	    case gl.QUADS:    // no longer exists 
		break;
	    case gl.LINE_LOOP:
		// if (gl1.INDEXED) gl1.indexArray.push(gl1.first, gl1.second);
		break;
	    }

	    
	    // --- bind buffers to shader attributes & draw

	    if (gl1.shadeModel==gl.SMOOTH) // not compatible with display-list
		gl1.drawVBO();
	    else
		gl1.drawFlat();

	    // gl1.inDraw = false; done in drawVBO

	    // the real finish of end() is at the end of drawVBO
	};


	// flatten, convert to float32 and send to last binded buffer.
	    // TODO: would be better to avoid such array casts
	gl1.setBuffer = function( B ) {
	    if (gl1.proto.callList) return; // replay mode: buffer already defined
	    gl.bufferData(gl.ARRAY_BUFFER,
			  new Float32Array([].concat.apply([],B)),
			   gl1.VBOhint);
	}
	
	// bind buffers to shader attributes & draw
	gl1.drawVBO = function() {
	    if (gl1.index==0) {  gl1.inDraw = false; return; } // empty block
	    
	    gl1.VBOhint = gl.STATIC_DRAW;
	    //gl1.VBOhint = gl.DYNAMIC_DRAW;
	    //gl1.VBOhint = gl.STREAM_DRAW;

	    // --- bind buffers and connect to shader attributes
	    // Maybe MV & NMV transformed (in displayLists) should be done here

	    if (gl1.proto.inList || !gl1.vertexObject)
		gl1.vertexObject = gl.createBuffer();
	    gl.bindBuffer(gl.ARRAY_BUFFER, gl1.vertexObject);
	    gl1.setBuffer(gl1.vertexArray); // do nothing in lists
	    gl.vertexAttribPointer(gl1.shaderProgram.vertexPositionAttribute, 3, gl.FLOAT, false, 0, 0);
	    
	    if (gl1.normalEnabled) 
		if (gl1.normalSet) {
		    if (gl1.proto.inList || !gl1.normalObject)
			gl1.normalObject = gl.createBuffer();
		    gl.bindBuffer(gl.ARRAY_BUFFER, gl1.normalObject);
		    gl1.setBuffer(gl1.normalArray);
		    gl.vertexAttribPointer(gl1.shaderProgram.normalAttribute, 3, gl.FLOAT, false, 0, 0);
		    gl.enableVertexAttribArray(gl1.shaderProgram.normalAttribute);
		}
	    else {
		gl.vertexAttrib3f(gl1.shaderProgram.normalAttribute,gl1.nx,gl1.ny,gl1.nz);
		gl.disableVertexAttribArray(gl1.shaderProgram.normalAttribute);
	    }
	    
	    if (gl1.texCoordEnabled)
		if (gl1.texCoordSet) {
		    if (gl1.proto.inList || !gl1.texCoordObject) gl1.texCoordObject = gl.createBuffer();
		    gl.bindBuffer(gl.ARRAY_BUFFER, gl1.texCoordObject);
		    gl1.setBuffer(gl1.texCoordArray);
		    gl.vertexAttribPointer(gl1.shaderProgram.texCoordAttribute, 2, gl.FLOAT, false, 0, 0);
		    gl.enableVertexAttribArray(gl1.shaderProgram.texCoordAttribute);
		}
	    else {
		gl.vertexAttrib2f(gl1.shaderProgram.texCoordAttribute,gl1.u,gl1.v);
		gl.disableVertexAttribArray(gl1.shaderProgram.texCoordAttribute);
	    }
		
	    if (gl1.colorEnabled)
		if (gl1.colorSet) {
		    if (gl1.proto.inList || !gl1.colorObject) gl1.colorObject = gl.createBuffer();
		    gl.bindBuffer(gl.ARRAY_BUFFER, gl1.colorObject);
		    gl1.setBuffer(gl1.colorArray);
		    gl.vertexAttribPointer(gl1.shaderProgram.colorAttribute, 4, gl.FLOAT, false, 0, 0);
		    gl.enableVertexAttribArray(gl1.shaderProgram.colorAttribute);
	    }
	    else {
		gl.vertexAttrib4f(gl1.shaderProgram.colorAttribute,gl1.r,gl1.g,gl1.b,gl1.a);
		gl.disableVertexAttribArray(gl1.shaderProgram.colorAttribute);
	    }
	    // limited support of back-side color
	    if (gl1.shaderProgram.colorBAttribute>=0)
		gl.vertexAttrib4f(gl1.shaderProgram.colorBAttribute,gl1.rB,gl1.gB,gl1.bB,gl1.aB);
	
	    if (gl1.INDEXED) { // if displayed as indexArray
		if (gl1.proto.inList || !gl1.indexObject) gl1.indexObject = gl.createBuffer();
		gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, gl1.indexObject);
		gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, new Uint16Array(gl1.indexArray),  gl1.VBOhint);
		
		gl1.numIndices = gl1.indexArray.length;
	    }
	    else gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, null);

	    
	    // --- draw
	    

	    if (!gl1.proto.inList) // immediate draw
		gl1.draw();
	    else {        // keep all buffers and states, defered to callList
		gl1.proto.elements.push(gl1);
		gl1.proto.nbElements++;
	    }
	    
	    gl1.inDraw = false;
	}
	
	gl1.draw = function() { // draw buffers (never called in display-lists)
	    
	    gl1.setUniforms(); // applies matrices, materials, fog...

	    var type = gl1.type;
	    if (gl1.proto.ctx.polygonMode == gl.LINE)
		type = gl.LINE_STRIP; // approx: should loop
	        // will do poorly of GL_TRIANGLES (Quads have been subdivided)
	    
	    if (gl1.INDEXED) // // if displayed as indexArray 
		// TODO: emulate for QUADS
		gl.drawElements(type, gl1.numIndices, gl.UNSIGNED_SHORT, 0);
	    else
		if (gl1.type != gl.QUADS) 
		    gl.drawArrays(type, 0, gl1.index);
	        else  {                           // emulation for QUADS
		    type = gl.TRIANGLE_FAN;
		    if (gl1.proto.ctx.polygonMode == gl.LINE)
			type = gl.LINE_STRIP;
		    for (var i = 0; i<gl1.index; i+= 4)
			gl.drawArrays(type, i, 4);	
		}
	}
	
	gl1.drawFlat = function () { // remake a VBO in case of flat shading
	    var vertex     = gl1.vertexArray;
	          normal   = gl1.normalArray,
	          color    = gl1.colorArray,
                  texCoord = gl1.texCoordArray,
	          index    = gl1.indexArray;
	    
	    switch (gl1.type) {
	    case gl.POINTS:
	    case gl.LINES:
	    case gl.LINE_STRIP:
	    case gl.LINE_LOOP:
                                //  bypass flattening for 0D & 1D shapes. TODO.
		gl1.INDEXED = false;
		gl1.drawVBO();
		// gl1.inDraw = false; done in drawVBO
		break;

	    case gl.TRIANGLES:
	    case gl.TRIANGLE_STRIP:
	    case gl.TRIANGLE_FAN:
	    case gl.QUADS:   
	    case gl.QUAD_STRIP:
	    case gl.POLYGON: 
		                // all 2D shapes have been triangulated
		normalSet=gl1.normalSet; colorSet=gl1.colorSet; texCoordSet=gl1.texCoordSet;
		gl.begin(gl.TRIANGLES);
		gl1.normalSet=normalSet; gl1.colorSet=colorSet; gl1.texCoordSet=texCoordSet;
		gl1.INDEXED = false; // was set by first plain draw
		for (var i=0; i<index.length/3; i++) {
		    //var j0 = index[i][0], j1 = index[i][1], j2 = index[i][2];
		    var j0 = index[3*i], j1 = index[3*i+1], j2 = index[3*i+2];
		    if (gl1.normalEnabled && gl1.normalSet) {
			var nx = normal[j0][0] + normal[j1][0] + normal[j2][0],
		        ny = normal[j0][1] + normal[j1][1] + normal[j2][1],
		        nz = normal[j0][2] + normal[j1][2] + normal[j2][2];
			gl.normal3f(nx/3,ny/3,nz/3);
		    }
		    if (gl1.colorEnabled && gl1.colorSet) {
			var r = color[j0][0] + color[j1][0] + color[j2][0],
		            g = color[j0][1] + color[j1][1] + color[j2][1],
		            b = color[j0][2] + color[j1][2] + color[j2][2],
		            a = color[j0][3] + color[j1][3] + color[j2][3];
			gl.color4f(r/3,g/3,b/3,a/3);
		    }
		    if (gl1.texCoordEnabled&& gl1.texCoordSet) { // correct ?
			var u = texCoord[j0][0] + texCoord[j1][0] + texCoord[j2][0],
		        v = texCoord[j0][1] + texCoord[j1][1] + texCoord[j2][1];
			gl.texCoord2f(u/3,v/3);
		    }
		    gl.vertex3fv(vertex[j0]);
		    gl.vertex3fv(vertex[j1]);
		    gl.vertex3fv(vertex[j2]); // only case with no gl.end
		}
	    
		gl1.drawVBO(); // gl.end would loop ;-)
		// gl1.inDraw = false; done in drawVBO
		break;
	    }
	    
	}

	
	// --- glu solids
	
	glu.newQuadric = function() {}

	glu.sphere = function(quad, radius, nbLong, nbLat) {
	    
	    gl.begin(gl.QUAD_STRIP);
	    
	    for (var j = 0; j <= nbLat; j++) {
		var theta = j * Math.PI / nbLat,
		    sinTheta = Math.sin(theta),
		    cosTheta = Math.cos(theta);

		for (var i = 0; i <= nbLong; i++) {
		    var phi = i * 2 * Math.PI / nbLong,
		        sinPhi = Math.sin(phi),
		        cosPhi = Math.cos(phi);
		    
		    var x = cosPhi * sinTheta,
		        y = cosTheta,
		        z = sinPhi * sinTheta;
		    
		    gl.normal3f(x,y,z);
		    gl.texCoord2f( 1 - i/nbLong, 1 - j/nbLat);
		    gl.vertex3f(radius * x, radius * y, radius * z);
		}
	    }
	    
	    for (var j = 0; j < nbLat; j++) {
		for (var i = 0; i < nbLong; i++) {
		    var first = (j * (nbLong + 1)) + i;
		    var second = first + nbLong + 1;
		    gl1.indexArray.push(first);
		    gl1.indexArray.push(second);
		    gl1.indexArray.push(first + 1);
		    
		    gl1.indexArray.push(second);
		    gl1.indexArray.push(second + 1);
		    gl1.indexArray.push(first + 1);
		}
	    }

	    gl1.INDEXED = true;
	    gl.end();
	    gl1.INDEXED = false;
	}
	
	glu.disk = function(quad, r0, r1, nbLong, nbLat) {
	    
	    gl.begin(gl.TRIANGLES);
	    
	    for (var j = 0; j <= nbLat; j++) {
		var radius = r0 + j * (r1-r0) / nbLat;

		for (var i = 0; i <= nbLong; i++) {
		    var phi = i * 2 * Math.PI / nbLong,
		        sinPhi = Math.sin(phi),
		        cosPhi = Math.cos(phi);
		    
		    var x =  cosPhi,
		        y = -sinPhi,
		        z = 0;
		    
		    gl.normal3f(0,0,1);
		    gl.texCoord2f( 1 - i/nbLong, 1 - j/nbLat);
		    gl.vertex3f(radius * x, radius * y, radius * z);
		}
	    }
	    
	    for (var j = 0; j < nbLat; j++) {
		for (var i = 0; i < nbLong; i++) {
		    var first = (j * (nbLong + 1)) + i;
		    var second = first + nbLong + 1;
		    gl1.indexArray.push(first);
		    gl1.indexArray.push(second);
		    gl1.indexArray.push(first + 1);
		    
		    gl1.indexArray.push(second);
		    gl1.indexArray.push(second + 1);
		    gl1.indexArray.push(first + 1);
		}
	    }

	    gl1.INDEXED = true;
	    gl.end();
	    gl1.INDEXED = false;
	}
	
	glu.cylinder = function(quad, r0, r1, h, nbLong, nbLat) {
	    
	    gl.begin(gl.TRIANGLES);
	    
	    for (var j = 0; j <= nbLat; j++) {
		var z = j * h / nbLat,
		    radius = r0 + j * (r1-r0) / nbLat;

		for (var i = 0; i <= nbLong; i++) {
		    var phi = i * 2 * Math.PI / nbLong,
		        sinPhi = Math.sin(phi),
		        cosPhi = Math.cos(phi);
		    
		    var x = cosPhi,
		        y = sinPhi;
		    
		    gl.normal3f(x,y,0);
		    gl.texCoord2f( 1 - i/nbLong, 1 - j/nbLat);
		    gl.vertex3f(radius * x, radius * y, h * z);
		}
	    }
	    
	    for (var j = 0; j < nbLat; j++) {
		for (var i = 0; i < nbLong; i++) {
		    var first = (j * (nbLong + 1)) + i;
		    var second = first + nbLong + 1;
		    gl1.indexArray.push(first);
		    gl1.indexArray.push(second);
		    gl1.indexArray.push(first + 1);
		    
		    gl1.indexArray.push(second);
		    gl1.indexArray.push(second + 1);
		    gl1.indexArray.push(first + 1);
		}
	    }

	    gl1.INDEXED = true;
	    gl.end();
	    gl1.INDEXED = false;
	}

	gl1.torus = function(r0,r1, nbLong, nbLat) {
	    var R = (r0+r1)/2, r = Math.abs(r1-r0)/2;
	    
	    gl.begin(gl.QUAD_STRIP);
	    
	    for (var j = 0; j <= nbLat; j++) {
		var theta = j * 2. * Math.PI / nbLat,
		    sinTheta = Math.sin(theta),
		    cosTheta = Math.cos(theta);

		for (var i = 0; i <= nbLong; i++) {
		    var phi = i * 2 * Math.PI / nbLong,
		        sinPhi = Math.sin(phi),
		        cosPhi = Math.cos(phi);
		    
		    var x = cosPhi * (R + r*cosTheta),
	                y = sinPhi * (R + r*cosTheta),
		        z =               r*sinTheta;
		    
		    gl.normal3f(cosPhi*cosTheta, cosPhi*sinTheta, sinTheta);
		    gl.texCoord2f( 1 - i/nbLong, 1 - j/nbLat);
		    gl.vertex3f(x,y,z);
		}
	    }
	    
	    for (var j = 0; j < nbLat; j++) {
		for (var i = 0; i < nbLong; i++) {
		    var first = (j * (nbLong + 1)) + i;
		    var second = first + nbLong + 1;
		    gl1.indexArray.push(first);
		    gl1.indexArray.push(second);
		    gl1.indexArray.push(first + 1);
		    
		    gl1.indexArray.push(second);
		    gl1.indexArray.push(second + 1);
		    gl1.indexArray.push(first + 1);
		}
	    }

	    gl1.INDEXED = true;
	    gl.end();
	    gl1.INDEXED = false;
	}

	gl1.cube = function(s) { // orientation issues
	    s /= 2.;
	    gl.begin(gl.QUADS);
	    gl.normal3f(0,0,-1);
	    gl.vertex3f(-s,-s,-s);
	    gl.vertex3f( s,-s,-s);
	    gl.vertex3f( s, s,-s);
	    gl.vertex3f(-s, s,-s);
	    gl.normal3f(0,0,1);
	    gl.vertex3f(-s,-s, s);
	    gl.vertex3f( s,-s, s);
	    gl.vertex3f( s, s, s);
	    gl.vertex3f(-s, s, s);
	    gl.normal3f(0,-1,0);
	    gl.vertex3f(-s,-s,-s);
	    gl.vertex3f( s,-s,-s);
	    gl.vertex3f( s,-s, s);
	    gl.vertex3f(-s,-s, s);
	    gl.normal3f(1,0,0);
	    gl.vertex3f( s,-s,-s);
	    gl.vertex3f( s, s,-s);
	    gl.vertex3f( s, s, s);
	    gl.vertex3f( s,-s, s);
	    gl.normal3f(0,1,0);
	    gl.vertex3f( s, s,-s);
	    gl.vertex3f(-s, s,-s);
	    gl.vertex3f(-s, s, s);
	    gl.vertex3f( s, s, s);
	    gl.normal3f(-1,0,0);
	    gl.vertex3f(-s, s,-s);
	    gl.vertex3f(-s,-s,-s);
	    gl.vertex3f(-s,-s, s);
	    gl.vertex3f(-s, s, s);

	    //gl1.indexArray.push(0,1,2,3,  7,6,5,4);  // no QUADS !
	    //gl1.indexArray.push(0,1,5,4, 1,2,6,5, 2,3,7,6, 3,0,4,7);
	    //gl1.indexArray.push(0,1,3,1,2,3,  7,6,4,6,5,4); // separate N !
	    //gl1.indexArray.push(0,1,4,1,5,4, 1,2,5,2,6,5, 2,3,6,3,7,6, 3,0,7,0,4,7);

	    gl.end();
	}

	gl1.tetrahedron = function() {
	    gl1.vertexArray = [
		[  1.714, -0.606,  0    ],
		[ -0.857, -0.606, -1.485],
		[ -0.857, -0.606,  1.485],
	        [  0     , 1.818,  0    ] ];
	    gl1.normalArray = gl1.vertexArray;
	    //var col = [gl1.r, gl1.g, gl1.b];
	    //gl1.colorArray = [ col,col,col,col];
	    gl1.indexArray = [  0, 1, 2,   3, 1, 0,   3, 0, 2,   3, 2, 1 ];
	    gl1.type = gl.TRIANGLES;
	    gl1.normalSet = true;
	    gl1.drawFlat();
	}	
	
	gl1.icosahedron = function() {
	    gl1.vertexArray = [ 
		[ -0, -0.525731, 0.850651  ],
		[ -0.850651, 0, 0.525731   ],     
		[ -0.850651, 0, -0.525731  ],    
		[  0.850651, 0, -0.525731  ],  
		[  0.850651, 0, 0.525731   ], 
		[  0.525731, 0.850651, 0   ],     
		[ -0.525731, 0.850651, 0   ],      
		[ -0.525731, -0.850651, 0  ],      
		[  0.525731, -0.850651, 0  ],     
		[ -0, -0.525731, -0.850651 ],      
		[ -0, 0.525731, -0.850651  ],      
		[ -0, 0.525731, 0.850651   ] ];
	    gl1.normalArray = [ 
		[ -0.000000, -0.417775, 0.675974  ],
		[ -0.675973, 0.000000, 0.417775   ],
		[ -0.675973, -0.000000, -0.417775 ],
		[  0.675973, 0.000000, -0.417775  ],
		[  0.675973, -0.000000, 0.417775  ],
		[  0.417775, 0.675974, 0.000000   ],
		[ -0.417775, 0.675973, -0.000000  ],
		[ -0.417775, -0.675974, 0.000000  ],
		[  0.417775, -0.675974, 0.000000  ],
		[ -0.000000, -0.417775, -0.675973 ],
		[ -0.000000, 0.417775, -0.675974  ],
		[ -0.000000, 0.417775, 0.675973   ]   ];
	    gl1.indexArray = [ 
		 1, 2, 6 ,    1, 7, 2 ,   3,  4,  5,    4,  3, 8,
		 6, 5, 11,    5, 6, 10,   9, 10,  2,   10,  9, 3,
		 7, 8, 9 ,    8, 7, 0 ,   11, 0,  1,    0, 11, 4,
		 6, 2, 10,    1, 6, 11,    3, 5, 10,    5, 4, 11,
		 2, 7, 9 ,    7, 1, 0 ,    3, 9,  8,    4, 8,  0   ];
	    gl1.normalSet = true;
	    gl1.type = gl.TRIANGLES;
	    gl1.drawFlat();
	}
	
	glut.solidTorus = function(r0,r1, nbLong, nbLat) {
	    var prev_mode = gl1.polygonMode;
	    gl.polygonMode(gl.FRONT_AND_BACK, gl.FILL);
	    gl1.torus(r0,r1, nbLong, nbLat);
	    gl.polygonMode(gl.FRONT_AND_BACK, prev_mode);
	}
	glut.wireTorus = function(r0,r1, nbLong, nbLat) {
	    var prev_mode = gl1.polygonMode;
	    gl.polygonMode(gl.FRONT_AND_BACK, gl.LINE);
	    gl1.torus(r0,r1, nbLong, nbLat);
	    gl.polygonMode(gl.FRONT_AND_BACK, prev_mode);
	}
	
	glut.solidCube = function(s) {
	    var prev_mode = gl1.polygonMode;
	    gl.polygonMode(gl.FRONT_AND_BACK, gl.FILL);
	    gl1.cube(s);
	    gl.polygonMode(gl.FRONT_AND_BACK, prev_mode);
	}	
	glut.wireCube = function(s) {
	    var prev_mode = gl1.polygonMode;
	    gl.polygonMode(gl.FRONT_AND_BACK, gl.LINE);
	    gl1.cube(s);
	    gl.polygonMode(gl.FRONT_AND_BACK, prev_mode);
	}

	glut.solidSphere = function( radius, nbLong, nbLat) {
	    var prev_mode = gl1.polygonMode;
	    gl.polygonMode(gl.FRONT_AND_BACK, gl.FILL);
	    glu.sphere(0, radius, nbLong, nbLat);
	    gl.polygonMode(gl.FRONT_AND_BACK, prev_mode);
	}	
	glut.wireSphere = function( radius, nbLong, nbLat) {
	    var prev_mode = gl1.polygonMode;
	    gl.polygonMode(gl.FRONT_AND_BACK, gl.LINE);
	    glu.sphere(0, radius, nbLong, nbLat);
	    gl.polygonMode(gl.FRONT_AND_BACK, prev_mode);
	}
	
	glut.solidCone = function (r0, h, nbLong, nbLat) {
	    var prev_mode = gl1.polygonMode;
	    gl.polygonMode(gl.FRONT_AND_BACK,gl.FILL);
	    glu.cylinder(0, r0, 0, h, nbLong, nbLat);
	    gl.polygonMode(gl.FRONT_AND_BACK, prev_mode);
	}
	glut.wireCone = function (r0, h, nbLong, nbLat) {
	    var prev_mode = gl1.polygonMode;
	    gl.polygonMode(gl.FRONT_AND_BACK,gl.LINE);
	    glu.cylinder(0, r0, 0, h, nbLong, nbLat);
	    gl.polygonMode(gl.FRONT_AND_BACK, prev_mode);
	}
	
	glut.solidTetrahedron = function () {
	    var prev_mode = gl1.polygonMode;
	    gl.polygonMode(gl.FRONT_AND_BACK,gl.FILL);
	    gl1.tetrahedron();
	    gl.polygonMode(gl.FRONT_AND_BACK, prev_mode);
	}
	glut.wireTetrahedron = function () {
	    var prev_mode = gl1.polygonMode;
	    gl.polygonMode(gl.FRONT_AND_BACK,gl.LINE);
	    gl1.tetrahedron();
	    gl.polygonMode(gl.FRONT_AND_BACK, prev_mode);
	}

	glut.solidIcosahedron = function () {
	    var prev_mode = gl1.polygonMode;
	    gl.polygonMode(gl.FRONT_AND_BACK,gl.FILL);
	    gl1.icosahedron();
	    gl.polygonMode(gl.FRONT_AND_BACK, prev_mode);
	}
	glut.wireIcosahedron = function () {
	    var prev_mode = gl1.polygonMode;
	    gl.polygonMode(gl.FRONT_AND_BACK,gl.LINE);
	    gl1.icosahedron();
	    gl.polygonMode(gl.FRONT_AND_BACK, prev_mode);
	}
    }



    
// --- emulates OpenGL1 matrices operations --------------------

    function initMatrices() {
	// var glMatrix = require('./gl-matrix');
	var mat4 = glMatrix.mat4,
	    vec4 = glMatrix.vec4,
	    vec3 = glMatrix.vec3;

	var mvMatrix = { val: mat4.create() }; // modelView
	var pMatrix  = { val: mat4.create() }; // projection
	var tMatrix  = { val: mat4.create() }; // texture
	var cMatrix  = { val: mat4.create() }; // color
	var MatrixStack = [];
	var nMatrix = { val: mat4.create() }; // inv transp of modelView
	var curMatrix = pMatrix; // default value ?

	gl1.calcNMatrix = function() {
	    nMatrix.val = mat4.clone(mvMatrix.val);
	    mat4.invert(nMatrix.val, nMatrix.val);
	    mat4.transpose(nMatrix.val, nMatrix.val);
	}
	gl1.getMVMatrix = function() { return mvMatrix.val; }
	gl1.getPMatrix  = function() { return pMatrix.val;  }
	gl1.getNMatrix  = function() { return NMatrix.val;  }
	
	function setUniform4f(target, v) { if (target) gl.uniform4f(target, v[0],v[1],v[2],v[3]); }
	function setUniform1f(target, v) { if (target) gl.uniform1f(target, v);                   }
	function setUniform1i(target, v) { if (target) gl.uniform1i(target, v);                   }

	gl1.setUniforms = function() { // plug matrices to shader
	    // matrices
	    gl.uniformMatrix4fv(gl1.shaderProgram.pMatrixUniform,  false, pMatrix.val );
	    gl.uniformMatrix4fv(gl1.shaderProgram.mvMatrixUniform, false, mvMatrix.val);
	    gl1.calcNMatrix();
	    gl.uniformMatrix4fv(gl1.shaderProgram.nMatrixUniform, false, nMatrix.val);
	    // lights
	    setUniform1i(gl1.shaderProgram.lightingEnabled, gl1.proto.ctx.lightingEnabled ?1:0);
	    if (gl1.proto.ctx.lightingEnabled) { // is it wise ?
	       //gl.uniform4fv(gl1.shaderProgram.lightDirUniform, new Float32Array(gl1.lightPos0));  
	       //gl.uniform4fv(gl1.shaderProgram.lightDirUniform, gl1.lightPos0);	
		setUniform4f(gl1.shaderProgram.light0PositionUniform, gl1.proto.ctx.light0Position);
		setUniform4f(gl1.shaderProgram.light0AmbientUniform,  gl1.proto.ctx.light0Ambient);
		setUniform4f(gl1.shaderProgram.light0DiffuseUniform,  gl1.proto.ctx.light0Diffuse);
		setUniform4f(gl1.shaderProgram.light0SpecularUniform, gl1.proto.ctx.light0Specular);
		setUniform4f(gl1.shaderProgram.materialSpecularUniform, gl1.proto.ctx.materialSpecular);
		setUniform1f(gl1.shaderProgram.materialShininessUniform, gl1.proto.ctx.materialShininess);
		setUniform4f(gl1.shaderProgram.materialEmissionUniform, gl1.proto.ctx.materialEmission);
	    }

	    // fog
	    setUniform1i(gl1.shaderProgram.fogEnabled, gl1.proto.ctx.fogEnabled ?1:0);
	    if (gl1.proto.ctx.fogEnabled) {
		setUniform4f(gl1.shaderProgram.fogColorUniform, gl1.proto.ctx.fogColor);
		setUniform1f(gl1.shaderProgram.fogDensityUniform, gl1.proto.ctx.fogDensity);
	    }

	    // texture
	    setUniform1i(gl1.shaderProgram.textureEnvModeUniform, (gl1.texEnabled && gl.getParameter(gl.TEXTURE_BINDING_2D)) ? gl1.TEXTURE_ENV_MODE : 0);
	    setUniform1i(gl1.shaderProgram.textureSamplerUniform, 0); // texture0 - gl.activeTexture(gl.TEXTURE0) = default

	    // misc
	    setUniform1f(gl1.shaderProgram.pointSizeUniform, gl1.proto.ctx.pointSize);
	    
	}
	
	gl1.applyMV = function(v) {
	    v = vec4.clone(v);
	    //var tmatrix = mat4.clone(mvMatrix.val);
	    //mat4.transpose(tMatrix,tMatrix);
	    v = vec4.transformMat4(v,v,mvMatrix.val);
	    return v; //vec4.normalize(v,v);
	}
	gl1.applyInvMV = function(v) {
	    v = vec4.clone(v);
	    var imatrix = mat4.clone(mvMatrix.val);
	    mat4.invert(iMatrix,iMatrix);
	    v = vec4.transformMat4(v,v,iMatrix.val);
	    return v; 
	}
	gl1.applyNMV = function(n) {
	    n = vec4.clone(n);
	    n = vec4.transformMat4(n,n,nMatrix.val);
	    return n; 
	}
    
	// Matrix Mode --- from gl.h
	gl.MATRIX_MODE     = 0x0BA0;
	gl.MODELVIEW       = 0x1700;
	gl.PROJECTION      = 0x1701;
	gl.TEXTURE	   = 0x1702;
	gl.COLOR	   = 0x1800;

	gl.matrixMode = function(cur_matrix) { 
	    switch(cur_matrix) {
	    case gl.PROJECTION: curMatrix = pMatrix;  break; 
	    case gl.MODELVIEW:  curMatrix = mvMatrix; break;
	    case  gl.TEXTURE:   curMatrix = tMatrix;  break
	    case  gl.COLOR:     curMatrix = cMatrix;  break;
	    }
	}
	
	gl.pushMatrix = function() {
	    MatrixStack.push(mat4.clone(curMatrix.val));
	};
	gl.popMatrix  = function() {
	    mat4.copy(curMatrix.val, MatrixStack.pop());
	};    
	gl.loadIdentity = function() {
	    mat4.identity(curMatrix.val);
	};
	gl.translatef = function(x,y,z) {
	    mat4.translate(curMatrix.val,curMatrix.val,[x,y,z]);
	};
	gl.scalef = function(x,y,z) {
	    mat4.scale(curMatrix.val,curMatrix.val,[x,y,z]);
	};
	gl.rotatef = function(deg, x,y,z) {
	    mat4.rotate(curMatrix.val,curMatrix.val,deg/180*Math.PI, [x,y,z]);
	};	
	gl.multMatrixf  = function(m) {
	    mat4.multiply(curMatrix.val,curMatrix.val,m);
	};	
	gl.ortho = function(x0,x1,y0,y1,z0,z1) {
	    mat4.ortho(curMatrix.val, x0,x1, y0,y1,  z0,z1);
	};
	gl.frustum = function(x0,x1,y0,y1,z0,z1) {
	    mat4.frustum(curMatrix.val, x0,x1, y0,y1,  z0,z1);
	}
	glu.perspective = function(fovy,aspect,zNear,zFar) {
	    mat4.perspective(curMatrix.val, fovy/180*Math.PI,aspect,zNear,zFar);
	    /* // for debug picking
	       if (gl1.pickMatrix) { 
		var x = gl1.pickViewport[0], y = gl1.pickViewport[1],
	            w = gl1.pickViewport[2], h = gl1.pickViewport[3];
		gl.viewport(x-w/2,y-h/2,w,h);
		gl1.applyPickMatrix(gl1.pickMatrix);
	    }*/
	};
	glu.lookAt = function(ex,ey,ez, cx,cy,cz, ux,uy,uz) { // eye, center, up
	    mat4.lookAt(curMatrix.val, [ex,ey,ez], [cx,cy,cz], [ux,uy,uz]);
	}
	gl1.applyPickMatrix = function(m) {
	    mat4.multiply(pMatrix.val,m,pMatrix.val);
	    //mat4.multiply(pMatrix.m,pMatrix.val,m);
	}

	// from https://nickthecoder.wordpress.com/2013/01/17/unproject-vec3-in-gl-matrix-library/	
	vec3.unproject = function (vec, view, proj, viewport) {
	    
	    var dest = vec3.create();//output
	    var m = mat4.create();//view * proj
	    var im = mat4.create();//inverse view proj
	    var v = vec4.create();//vector
	    var tv = vec4.create();//transformed vector
	    
	    //apply viewport transform
	    v[0] = (vec[0] - viewport[0]) * 2.0 / viewport[2] - 1.0;
	    v[1] = (vec[1] - viewport[1]) * 2.0 / viewport[3] - 1.0;
	    v[2] = vec[2];
	    v[3] = 1.0;
	    
	    //build and invert viewproj matrix
	    mat4.multiply(m,view,proj);
	    if(!mat4.invert(im,m)) { return null; }
	    
	    vec4.transformMat4(tv,v,im);
	    if(v[3] === 0.0) { return null; }
	    
	    dest[0] = tv[0] / tv[3];
	    dest[1] = tv[1] / tv[3];
	    dest[2] = tv[2] / tv[3];
	    
	    return dest;
	};
	
	glu.unProjectv = function(x,y,z, mvmatrix,pmatrix, viewport, P) {
	    vec3.copy(P, vec3.unproject( [x,y,z], mvmatrix,pmatrix, viewport));
	}

	glu.pickMatrix = function(x,y,w,h,vp) {
	    
	    var  out = mat4.create();
	    out[0]  = vp[2] / w;
	    out[4]  = 0;
	    out[8]  = 0;
	    out[12] = ( vp[2] + 2 * (vp[0] - x) ) / w;
	    out[1]  = 0;
	    out[5]  = vp[3] / h;
	    out[9]  = 0;
	    out[13] = ( vp[3] + 2 * (vp[1] - y) ) / h;
	    out[2]  = 0;
	    out[6]  = 0;
	    out[10] = 1;
	    out[14] = 0;
	    out[3]  = 0;
	    out[7]  = 0;
	    out[11] = 0;
	    out[15] = 1;
	    gl.multMatrixf(out);

	    gl1.pickViewport = [x,y,w,h];
	    gl1.pickMatrix = out;
	}
    }




    
// --- emulates OpenGL 1 lighting ops -------------------- TO DO !

    function initLighting() {
    
	// Lighting --- from gl.h 
	gl.LIGHTING                     = 0x0B50 ;
	gl.LIGHT0                       = 0x4000 ;
	gl.LIGHT1                       = 0x4001 ;
        gl.AMBIENT			= 0x1200 ;
        gl.DIFFUSE			= 0x1201 ;
        gl.SPECULAR			= 0x1202 ;
        gl.EMISSION			= 0x1600 ;
        gl.POSITION			= 0x1203 ;
        gl.AMBIENT_AND_DIFFUSE		= 0x1602 ;
        gl.SHADE_MODEL			= 0x0B54 ;
        gl.FLAT				= 0x1D00 ;
        gl.SMOOTH			= 0x1D01 ;
        gl.COLOR_MATERIAL		= 0x0B57 ;
        //gl.COLOR_MATERIAL_FACE	= 0x0B55 ;
        //gl.COLOR_MATERIAL_PARAMETER	= 0x0B56 ;
	// GL_NORMALIZE			= 0x0BA1 ;

	// Fog
	gl.FOG				= 0x0B60 ;
	gl.FOG_DENSITY                  = 0x0B62 ;
	gl.FOG_COLOR                    = 0x0B66 ;

	gl1.shadeModel           = gl.SMOOTH; // should go in ctx
	gl1.ctx.lightingEnabled  =  false;
	gl1.ctx.light0Ambient    = [.2,.2,.2,1.];
	gl1.ctx.light0Diffuse    = [1.,1.,1.,1.];
	gl1.ctx.light0Specular   = [1.,1.,1.,1.];
	gl1.ctx.materialAmbient  = [.2,.2,.2,1.];
	gl1.ctx.materialDiffuse  = [.8,.8,.8,1.];
	gl1.ctx.materialSpecular = [0.,0.,0.,1.];
	gl1.ctx.materialEmission = [0.,0.,0.,1.];
	gl1.ctx.shininess        = 0.;
	gl1.side = gl.FRONT_AND_BACK; // ctx ?

	gl1.ctx.fogEnabled       = false;
	gl1.ctx.fogDensity       = 1.;
	gl1.ctx.fogColor         = [0.,0.,0.,0.];
	
	gl.shadeModel    = function (mode) { gl1.shadeModel = mode;  };
	gl.colorMaterial = function (side, mode)  {
	    // limited support of 2-sided materials
	    gl1.side = side;
	    if (side==gl.FRONT_AND_BACK) gl1.rB=-1;
	};
	
	gl.lightfv = function (id, param, val) { // plug lighting to shader
	    switch (param) { // shouldn't we clone val ?
	    case gl.POSITION :	gl1.ctx.light0Position = gl1.applyMV(val); break;
	    case gl.DIFFUSE  :  gl1.ctx.light0Diffuse  = val;    break; 
	    case gl.AMBIENT  :  gl1.ctx.light0Ambient  = val;    break; 
	    case gl.SPECULAR :  gl1.ctx.light0Specular = val;    break;
	    };
	};
	gl.materialfv = function (side, param, val) {
	    switch (param) { // shouldn't we clone val ?
	    case gl.AMBIENT :   gl1.ctx.material0Ambient  = val; break;
	    case gl.DIFFUSE :   gl1.ctx.materialDiffuse   = val; gl.color4fv(val); break;
	    case gl.SPECULAR :  gl1.ctx.materialSpecular  = val; break;
	    case gl.SHININESS : gl1.ctx.materialShininess = val; break;
	    case gl.EMISSION :  gl1.ctx.materialEmission  = val; break;
	    }
	};
	gl.materialf = gl.materialfv;
	gl.materiali = gl.materialfv;

	gl.lightModelf  = function(flag, value) {};
	gl.lightModeli  = function(flag, value) {};
	gl.lightModelfv = function(flag, value) {};
	gl.lightModeliv = function(flag, value) {};

	gl.fogfv = function (param, value) { // basic support
	    switch (param) {
	    case gl.FOG_DENSITY: gl1.ctx.fogDensity = value; break;
	    case gl.FOG_COLOR :  gl1.ctx.fogColor = value.slice(0); break;
	    }
	};
	gl.fogf = gl.flogfv;


	// Point & Polygons mode
	gl.polygonMode     = function (flag,mode) { gl1.ctx.polygonMode = mode; };
	gl.pointSize       = function(s) {  gl1.ctx.pointSize = s; }
	
	// --- from gl.h
	gl.POINT            = 0x1B00;
	gl.LINE             = 0x1B01;
	gl.FILL             = 0x1B02;
	
	gl1.ctx.polygonMode = gl.FILL; // set in init_lighting
	gl1.ctx.pointSize   = 1.;

	
	// texture mapping -- from gl.h
        gl.TEXTURE_ENV			= 0x2300;

        gl.COMBINE_RGB			= 0x8571;
        gl.COMBINE_ALPHA		= 0x8572;
        gl.TEXTURE_ENV_MODE		= 0x2200;
        gl.ADD				= 0x0104;
        gl.MODULATE			= 0x2100;
        gl.DECAL			= 0x2101;
//      gl.BLEND			= 0x0BE2; // already defined
//      gl.REPLACE			= 0x1E01; // already defined
        gl.COMBINE			= 0x8570;

        gl.S				= 0; // 0x2000
        gl.T				= 1; // 0x2001
        gl.TEXTURE_GEN_S		= 0x0C60;
        gl.TEXTURE_GEN_T		= 0x0C61;
        //gl.TEXTURE_GEN_R		= 0x0C62;
        //gl.TEXTURE_GEN_Q		= 0x0C63;
        gl.TEXTURE_GEN_MODE		= 0x2500;
        gl.OBJECT_LINEAR		= 0x2401;
        gl.OBJECT_PLANE			= 0x2501;
        gl.EYE_LINEAR			= 0x2400;
        gl.EYE_PLANE			= 0x2502;

	gl.TEXTURE_1D = gl.TEXTURE_2D; // no longer exist !
	gl1.texEnabled = false;
	gl1.texGenMode = [0,0];
	gl1.texGenPlane = [];

	
	// gl.BindTexture MUST take ids coming from genTexture, not ints !
	gl.genTextures = function (nb, ids) { // ids = return vector of ids
	    for (var i=0; i<nb; i++) ids.push(gl.createTexture());
	};
		    
	// textures 2D
	
	( function(){ // redefine gl.texImage2D to take buffers 
	    var tex2D = gl.texImage2D;
	    gl.texDOMImage2D = gl.texImage2D; // rename old one
	    gl.texImage2D = function(target,lev,iformat,w,h,b,format,type,buff) {
		//tex2D(target,lev,iformat,format,type, testImage );
		tex2D.call(gl, target,lev,iformat,w,h,b,format,type,
		      new Uint8Array([].concat.apply([],buff)) );
	    }
	}) ();
	
	( function(){ // redefine gl.texSubImage2D to take buffers 
	    var tex2D = gl.texSubImage2D;
	    gl.texDOMSubImage2D = gl.texSubImage2D; // rename old one
	    gl.texSubImage2D = function(target,lev,iformat,w,h,b,format,type,buff) {
		//tex2D(target,lev,iformat,format,type, testImage );
		tex2D.call(gl, target,lev,iformat,w,h,b,format,type,
		      new Uint8Array([].concat.apply([],buff)) );
	    }
	}) ();

	glu.build2DMipmaps = function(target,iformat,w,h,format,type,buff){
 	    gl.texImage2D(target,0,iformat,w,h,0,format,type,buff);
	    gl.generateMipmap(gl.TEXTURE_2D);
	}

	// textures 1D (no longer exist)
	
	gl.texImage1D = function(target,lev,iformat,w,b,format,type,buff) {
	    gl.texImage2D(target,lev,iformat,w,1,b,format,type,buff);
	}
	glu.build1DMipmaps = function(target,iformat,w,format,type,buff){
 	    gl.texImage1D(target,0,iformat,w,0,format,type,buff);
	    gl.generateMipmap(gl.TEXTURE_2D);
	}

	// texture common
	
	gl.texEnv = function(target, flag, val) { // todo
	    // target = gl.TEXTURE_ENV_MODE
	    switch(flag) {
	    case gl.COMBINE_RGB:      gl1.COMBINE_RGB = val; break;
	    case gl.COMBINE_ALPHA:    gl1.COMBINE_ALPHA = val; break;
	    case gl.TEXTURE_ENV_MODE: gl1.TEXTURE_ENV_MODE = val; break; // BLEND not implemented
	    }
	}; 
	gl.texEnvi = gl.texEnv;
	gl.texEnvf = gl.texEnv;
	
	gl.texGen = function(coord, mode, params) { // minimal support
	    switch (mode) {
	    case gl.OBJECT_PLANE : gl1.texGenPlane[coord] = params; break;
	    case gl.EYE_PLANE    : gl1.texGenPlane[coord] = gl1.applyInvMV(params); break;
	    case gl.TEXTURE_GEN_MODE: gl1.texGenMode[coord] = params; break;
	 // case gl.EYE_PLANE    :
	    }
	}
	gl.texGeni = gl.texGen; gl.texGeniv = gl.texGen;
	gl.texGenf = gl.texGen; gl.texGenfv = gl.texGen;
	gl.texGend = gl.texGen; gl.texGendv = gl.texGen;
	
	
	
// --- extends gl.enable() & gl.disable() to no longer existing flags 

	gl.LINE_SMOOTH      = 0x0B20;
	
    ( function(){
	var enableStd = gl.enable;
	 gl.enable = function(flag) {
	    switch(flag) {
	    case gl.LIGHTING : gl1.ctx.lightingEnabled = true;
		 //gl1.initShaders(); // <<<
		break; 
	    case gl.LIGHT0 :   gl1.ctx.light0Enabled   = true; break;
	    case gl.FOG :      gl1.ctx.fogEnabled      = true; break;
	    case gl.COLOR_MATERIAL : break; // on by default;
	    case gl.LINE_SMOOTH : break;  // no longer exist :-(
	    case gl.TEXTURE_GEN_S: gl1.texGen = true; break; 
	    case gl.TEXTURE_GEN_T: gl1.texGen = true; break;
	    case gl.TEXTURE_2D: gl1.texEnabled = true; break; 
	    default: enableStd.call(gl, flag);
	    } }
	 })();
	
    ( function(){
	var disableStd = gl.disable;
	gl.disable = function(flag) {
	    switch(flag) {
	    case gl.LIGHTING : gl1.ctx.lightingEnabled = false;
		 //gl1.initShaders(); // <<<
		break;
	    case gl.LIGHT0 :   gl1.ctx.light0Enabled   = false; break;
	    case gl.FOG :      gl1.ctx.fogEnabled      = false; break;
	    // case gl.COLOR_MATERIAL : break; // on by default;
	    case gl.LINE_SMOOTH : break; // no longer exist :-(
	    case gl.TEXTURE_GEN_S: gl1.texGen = false; break; 
	    case gl.TEXTURE_GEN_T: gl1.texGen = false; break;
	    case gl.TEXTURE_2D: gl1.texEnabled = false; break; 
	    default: disableStd.call(gl, flag);
	    } }
	 })();

	// Gets --- from gl.h
	gl.MODELVIEW_MATRIX =	0x0BA6;
	gl.PROJECTION_MATRIX =	0x0BA7;
	gl.VIEWPORT =		0x0BA2;
	
	gl.get = function (flag, value) {
	    var mat4 = glMatrix.mat4;

	    switch(flag) {
	    case gl.MODELVIEW_MATRIX: mat4.copy(value, gl1.getMVMatrix()); break;
	    case gl.PROJECTION_MATRIX: mat4.copy(value, gl1.getPMatrix()); break;
	    case gl.VIEWPORT: value[0]=0; value[1]=0; value[2]=gl1.width; value[3]=gl1.height; break; 
	    case gl.LIGHTING: return gl1.ctx.lightingEnabled; break;
	    case gl.LIGHT0: return true; break; // <<
	    //case gl.LIGHT1: break;
	    case gl.AMBIENT_AND_DIFFUSE: return true; break; // <<
	    case gl.SHADE_MODEL: return gl1.shadeModel; break;
  	    }
	}
	gl.getBooleanv = gl.get;
	gl.getDoublev = gl.get;
	gl.getFloatv = gl.get;
	gl.getIntegerv = gl.get;
	
    }
    
// --- emulates OpenGL 1 , glu , glut interface ops --------------------

    function initInterface(canvas) {
        
    // --- connect to canvas, setup resize and anim funcs (calling glut callbacks)  
    // NB: not compatible with multi-window
	
	var c = document.getElementById(canvas);
	if (!c)  return null;

	// force canvas size to window size. ugly.
	gl1.width  = c.width = window.innerWidth; // redundent with onresize (called at init)
	gl1.height = c.height = window.innerHeight;
	gl1.posx = 0; // ?
	gl1.posy = 0; // ?

	// there is a webGL alpha trap: http://webglfundamentals.org/webgl/lessons/webgl-and-alpha.html
	var options = {
	    // premultipliedAlpha: false  // Ask non-premultiplied alpha
	    // alpha: false
	    preserveDrawingBuffer: true
	};
	var gl = c.getContext('webgl',options) || c.getContext('experimental-webgl',options);
	
	
	// --- printf window
	
	// for more complete: http://phpjs.org/functions/sprintf/
	var stdout = document.getElementById("stdout");
	var stdout_show=1, stdout_trigger = null, stdout_rewind=false; 
	if (stdout)  stdout_trigger = stdout.getAttribute("triggerKey") || 'h';

	stdout_switch = function() {
	    stdout_show = 1-stdout_show;
	    stdout.style.visibility = (stdout_show) ? "visible" : "hidden";
	}

	sprintf = function(/*r,*/ f) { // arguments = format, var1..varn -> f, params[]
	    var params = Array.prototype.slice.call(arguments, 1); // 2
            // params.reduce(w(),f) : apply all params to format according to match function w()
	    // basic match: var w = function(p,c){return p.replace(/%./,c);}
	    // our matching is more complex:
	    var w = function(p,c){
		return p.replace(/([^%]?)%([0-9]*)(?:\.([0-9]+))?(c|s|d|f)/,
				 function(m,p,j,f,t) {
				     if ((t=='f') &&f) c = c.toFixed(f);
				     if (j) c=Array(Math.max(0,j-c.toString.length)+1).join(' ')+c;
				     if (p) c = p+c; // reinject leading char ( not-% )
				     return c;
				 });
	    }
	    f = params.reduce(w,f);
	    f = f.replace(/%%/g, "%");
	    //r.value = f;
	    return f;
	}

	printf = function() { // arguments = format, var1..varn -> f, params[]
	    //var params = Array.prototype.slice.call(arguments, 1);
	    var f = sprintf.apply(this,arguments);
	    // reformat spaces and returns, cancel last line if \r
	    f = f.replace(/\n/g, "<br>");
	    f = f.replace(/ /g, "&nbsp");
	    if  (stdout_rewind) stdout.innerHTML = stdout.innerHTML.slice(0,stdout.innerHTML.lastIndexOf("<br>"))+"<br>";
	    // display, scroll in necessary
	    stdout.innerHTML += f;
	    stdout.scrollTop = stdout.scrollHeight;
	    stdout_rewind = (f.slice(-1)=='\r');
	}

	fflush = function() {};

	// ---
	
	glut.swapBuffers   = function () {
	    // there is a webGL alpha trap: http://webglfundamentals.org/webgl/lessons/webgl-and-alpha.html
	    // -> Set the backbuffer's alpha to 1.0
	    gl.clearColor(1, 1, 1, 1);
	    gl.colorMask(false, false, false, true);
	    gl.clear(gl.COLOR_BUFFER_BIT);
	    gl.colorMask(true, true, true, true);
	}; 
	glut.mainLoop      = function () {}; // should not return
	glut.postRedisplay = function () {
	    if (glut.onDraw) window.requestAnimationFrame(glut.onDraw);
	};
	gl.drawBuffer      = function (b) {}; 
	gl.readBuffer      = function (b) {};
	
// --- from glut.h
// GLUT API macro definitions -- the glutGet parameters
	glut.WINDOW_X        = 0x0064;
	glut.WINDOW_Y        = 0x0065;
	glut.WINDOW_WIDTH    = 0x0066;
	glut.WINDOW_HEIGHT   = 0x0067;
	glut.WINDOW_STEREO   = 0x0079;

	glut.get = function (flag) {
	    switch(flag) {
	    case glut.WINDOW_X: return gl1.posx;
	    case glut.WINDOW_Y: return gl1.posy;
	    case glut.WINDOW_WIDTH:  return gl1.width;
	    case glut.WINDOW_HEIGHT: return gl1.height;
	    case glut.WINDOW_STEREO: return 0;
	    }
	}
	
	// --- resize and animate events
	
	glut.fullScreen    = function() {
            if( document.body.webkitRequestFullScreen ) {
		document.body.style.width = window.screen.width   + "px";
		document.body.style.height = window.screen.height + "px";
		document.body.webkitRequestFullScreen();
            }
            if( document.body.mozRequestFullScreen ) 
		document.body.mozRequestFullScreen();
            
	    window.onresize();
	}
	
	glut.reshapeWindow = function(w,h) {
	    gl1.width = w; 
	    gl1.height = h;
	    window.onresize()  ;      // ignore target size
	}
	glut.positionWindow = function(x,y) {}; // should do something

	window.onresize = function() {
	    // force canvas size to window size. ugly.
            gl1.width  = c.width  = window.innerWidth; // should be in gl1
            gl1.height = c.height = window.innerHeight;
            if (glut.onResize)	glut.onResize(gl1.width, gl1.height);
            if (glut.onDraw)    glut.onDraw();
	};
	
	var lastTime = 0;
        var vendors = ['ms', 'moz', 'webkit', 'o'];
        for (var x = 0; x < vendors.length && !window.requestAnimationFrame; ++x) {
	    window.requestAnimationFrame = window[vendors[x]+'RequestAnimationFrame'];
	    window.cancelAnimationFrame  = window[vendors[x]+'CancelAnimationFrame']
                                        || window[vendors[x]+'CancelRequestAnimationFrame'];
        }

	// utility for portable refresh func (used for anim or postredisplay)
        if (!window.requestAnimationFrame)
	    window.requestAnimationFrame = function(callback, element) {
                var currTime = new Date().getTime();
                var timeToCall = Math.max(0, 16 - (currTime - lastTime));
                var id = window.setTimeout(function() { callback(currTime + timeToCall); },
					   timeToCall);
                lastTime = currTime + timeToCall;
                return id;
	    };
	
        if (!window.cancelAnimationFrame)
	    window.cancelAnimationFrame = function(id) { clearTimeout(id); };


	// --- mouse & kbd events

	/*
	function stopDefAction(evt) { evt.preventDefault(); evt.stopPropagation(); }
	var d = window; // c / document / window
	d.addEventListener('click', stopDefAction, false);
	d.addEventListener('mousedown', stopDefAction, false);
	d.addEventListener('mouseup', stopDefAction, false);
	d.addEventListener( 'keypress', stopDefAction, false );
	d.addEventListener( 'keyup', stopDefAction, false );
	d.addEventListener( 'keydown', stopDefAction, false );
	*/
	
	// NB: mouse & kbd callback are defined even if user don't define glutCallbacks. not optimal.

	// attention: onkeypress see only printable character, no special keys.
	//            onkeydown see key, not if shifted (A=a, 1=&).
	newkey=true;
	document.onkeydown = function(e) { // =  keypress
	    //if (e.repeat||window.event.repeat) return;
	    e = e ? e : window.event;
	    var k = e.which || e.keyCode;
	    //c = String.fromCharCode(k); // not compatible with specials
	    if (gl1.onkeyS) gl1.onkeyS(k);
	}
	document.onkeypress = function(e) {
	    if (gl1.ignoreKeyRepeat && !newkey) return;
	    //if (e.repeat||window.event.repeat) return;
	    e = e ? e : window.event;
	    var k = e.which || e.keyCode;
	    if (stdout_trigger==String.fromCharCode(k)) stdout_switch(); 
	    //c = String.fromCharCode(k);
	    if (gl1.onkeyN) gl1.onkeyN(k);
	    newkey=false;
	}
	document.onkeyup = function(e) { 
	    //if (e.repeat||window.event.repeat) return;
	    e = e ? e : window.event;
	    var k = e.which || e.keyCode;
	    //c = String.fromCharCode(k); // not compatible with specials
	    if (gl1.onkeyU) gl1.onkeyU(k);
	    newkey=true;
	}

	glut.ignoreKeyRepeat = function(n) { gl1.ignoreKeyRepeat = (n>0) };
	
	
	var n = document.getElementById(canvas);
	var mousex,mousey, mousex_,mousey_, mousex0,mousey0, click; 
	mousex0 =  n.offsetLeft;
	mousey0 =  n.offsetTop;

	function validateMouse(mx,my) {
	    var valid=1;
	    if ((mx>=0) && (mx<gl1.width) ) mousex=mx; else valid=0;
	    if ((my>=0) && (my<gl1.height)) mousey=my; else valid=0;
	    
	    return valid;
	}
    
	document.onmousemove = function(e) { // = mousemove;
	//function mousemove(e){   // onmousemove
            e = e ? e : window.event;
            var mx = e.clientX - mousex0,
            my = e.clientY - mousey0;
	    if (validateMouse(mx,my) && gl1.mousemotion)
		gl1.mousemotion(mx,my);
	}
	
	function setModifiers(e) {
	    glut.shiftPressed = e.shiftKey;
	    glut.ctrlPressed  = e.ctrlKey;
	    glut.altPressed   = e.altKey;
	}
	
	glut.getModifiers = function() {
	    return   glut.ACTIVE_SHIFT* glut.shiftPressed
		   + glut.ACTIVE_CTRL * glut.ctrlPressed
		   + glut.ACTIVE_ALT  * glut.altPressed ;
	} 
	
	document.onmousedown = function(e) { // = mousedown;
	//function mousedown(e){  // onmousedown
	    e = e ? e : window.event;
	    setModifiers(e);
	    if (validateMouse(mousex,mousey) && gl1.mouseclick)
		gl1.mouseclick(e.button,glut.DOWN,mousex,mousey);
	}
    
	document.onmouseup   = function(e) { //= mouseup;
	//function mouseup(e)  {  // onmouseup
	    e = e ? e : window.event;
	    setModifiers(e);
	    if (validateMouse(mousex,mousey) && gl1.mouseclick)
		gl1.mouseclick(e.button,glut.UP,mousex,mousey);
	}

	// from gl.h 

	gl.TRUE  = true;   // webGL designers were purely bastards :
	gl.FALSE = false;  // this only undefined flag gl.TRUE was thus false.

	// from glut.h
	// GLUT API macro definitions -- the special key codes:
 
	glut.KEY_F1        = 112; // js keycodes // 0x0001;
	glut.KEY_F2        = 113; // 0x0002;
	glut.KEY_F3        = 114; // 0x0003;
	glut.KEY_F4        = 115; // 0x0004;
	glut.KEY_F5        = 116; // 0x0005;
	glut.KEY_F6        = 117; // 0x0006;
	glut.KEY_F7        = 118; // 0x0007;
	glut.KEY_F8        = 119; // 0x0008;
	glut.KEY_F9        = 120; // 0x0009;
	glut.KEY_F10       = 121; // 0x000A;
	glut.KEY_F11       = 122; // 0x000B;
	glut.KEY_F12       = 123; // 0x000C;
	glut.KEY_LEFT      =  37; // 0x0064;
	glut.KEY_UP        =  38; // 0x0065;
	glut.KEY_RIGHT     =  39; // 0x0066;
	glut.KEY_DOWN      =  40; // 0x0067;
	glut.KEY_PAGE_UP   =  33; // 0x0068;
	glut.KEY_PAGE_DOWN =  34; // 0x0069;
	glut.KEY_HOME      =  36; // 0x006A;
	glut.KEY_END       =  35; // 0x006B;
	glut.KEY_INSERT    =  45; // 0x006C;
	glut.KEY_NUM_LOCK  = 144; // 0x006D;
	glut.KEY_BEGIN     =         0x006E;
	glut.KEY_DELETE    =  46; // 0x006F;
	
	
	//  GLUT API macro definitions -- mouse state definitions
	
	glut.LEFT_BUTTON   = 0; // js event buttonL  // 0x0000;
	glut.MIDDLE_BUTTON = 1; // js event buttonM  // 0x0001;
	glut.RIGHT_BUTTON  = 2; // js event buttonR  // 0x0002;
	glut.DOWN          = 0x0000;
	glut.UP            = 0x0001;
	glut.LEFT          = 0x0000;
	glut.ENTERED       = 0x0001;
	
	// GLUT API macro definitions -- the glutGetModifiers parameters
	
	glut.ACTIVE_SHIFT  = 0x0001;
	glut.ACTIVE_CTRL   = 0x0002;
	glut.ACTIVE_ALT    = 0x0004;

	
	// --- glut menus 

	// require('./contextMenu.js');
	
	$('body').on('contextmenu', '#'+canvas, function(e){ return false; }); // switch off right button
	
	gl1.nbMenus = 0;
	gl1.menu = [];
	glut.createMenu = function(menu) {
	    gl1.curmenu = [];
	    gl1.menu.push(gl1.curmenu);
	    gl1.menuCallback = menu;
	    return gl1.nbMenus ++;
	};
	glutSetMenu = function(id) {
	    gl1.curmenu =  gl1.menu[id];
	};
	glut.addMenuEntry = function(label, val) {
	    gl1.curmenu.push( { name: label, fun: function() { gl1.menuCallback(val);} } );
	};
	glut.addSubMenu = function(label, id) {
	    gl1.curmenu.push( { name: label, subMenu: gl1.menu[id] } );
	}
	glut.attachMenu  = function(button) {
	    $('#'+canvas).contextMenu(gl1.curmenu,{triggerOn:'contextmenu'}); // right  button 
	    //$('#'+canvas).contextMenu(gl1.curmenu); // left+mid button

	};
	
	glut.changeToMenuEntry = function(id,label,val) {
	    gl1.curmenu[id-1] = { name: label, fun: function() { gl1.menuCallback(val);} };
	    glut.attachMenu();
	};
	
	glut.detachMenu = function(button) {
	    $('#'+canvas).contextMenu('destroy');
	};

	
	
	// --- glut fonts (only strokes Roman )
	
        //  all in require('./glutstrokefonts');
	
	
	// ------  glut callbacks -----------------------------------
	
	glut.idleFunc     = function(idle) {
	    glut.onIdle = idle;
	    window.requestAnimationFrame(onIdle0); // trigger first iteration
	}
	gl1.time = new Date().getTime();
	onIdle0 = function(idle) { // rearmed idle
	    window.requestAnimationFrame(onIdle0);
	    glut.onIdle();
	    // display fps
            var currTime = new Date().getTime();
	    var dt = (currTime-gl1.time)/1000.;
	    if (dt>1e-3) {
		gl1.time=currTime;
		var fpsDisplay = document.getElementById("fps");
		if (fpsDisplay) fpsDisplay.innerHTML=(1/dt).toFixed(1)+"fps";
	    }
	}
	glut.displayFunc  = function(redisplay) {
	    glut.onDraw   = redisplay;
	    glut.postRedisplay(); // arm first draw
	};
	glut.reshapeFunc  = function(reshape) {
	    glut.onResize = reshape;
	    window.onresize();
	    //reshape(gl1.width, gl1.height); // so as to set user size vars
	};
	glut.visibilityFunc = function(vis) { // TODO: to call somewhere !
	    glut.onVisible = vis;
	};
/*	doKey = function(k) { // map 2 glut functions in 1 html callback
	    if (glut.keyN) glut.keyN(String.fromCharCode(k),mousex,mousey);
	    if (glut.keyS) glut.keyS(k,mousex,mousey);
	}
	glut.keyboardFunc = function(key) {
	    glut.keyN = key;
	    gl1.onkey = doKey;
	};
	glut.specialFunc  = function(key) {
	    glut.keyS = key;
	    gl1.onkey = doKey;
	};
*/  /*
	 doKeyN = function(k) {
	    glut.keyN(String.fromCharCode(k),mousex,mousey);
	}
	doKeyS = function(k) {
	    glut.keyS(k,mousex,mousey);
	}
	doKeyU = function(k) {
	    glut.keyU(k,mousex,mousey);
	}*/
	glut.keyboardFunc = function(key) {
	    //glut.keyN = key;
	    gl1.onkeyN = function(k) { key(String.fromCharCode(k),mousex,mousey); }
	};
	glut.specialFunc  = function(key) {
	    // glut.keyS = key;
	    gl1.onkeyS = function(k) { key(k,mousex,mousey); }
	};
	glut.keyboardUpFunc = function(key) {
	    // glut.keyU = key;
	    gl1.onkeyU = function(k) {
		// very uggly and partial fix for this issue:
		// up/down event is about the physical key of the keyboard, not the character is generated depending on modifier :-(
		var c = String.fromCharCode(k);
		key(c,mousex,mousey);
		key(c.toLowerCase(),mousex,mousey);
	    }
	}
	glut.mouseFunc    = function(mouse_click)  {
	    gl1.mouseclick  = mouse_click;
	};
	glut.motionFunc   = function(mouse_motion) {
	    gl1.mousemotion = mouse_motion;
	};

	return gl;
    }
}